"""Tested function: dump."""

import ast
import logging
import unittest

import astunparse
import typed_ast.ast3
import typed_astunparse

from .examples import MODES, EXAMPLES, PATHS

_LOG = logging.getLogger(__name__)


def _postprocess_dump(tested_typed_dump):

    lines = [s.strip() for s in tested_typed_dump.splitlines()]
    lines = [s + ' ' if s.endswith(',') else s for s in lines]
    return ''.join(lines)


class DumpTests(unittest.TestCase):

    """Unit tests for dump() function."""

    maxDiff = None

    def test_dump_examples(self):
        """Print ASTs of examples correctly."""
        for description, example in EXAMPLES.items():
            for mode in MODES:
                if example['trees'][mode] is None:
                    continue
                dump = typed_astunparse.dump(example['trees'][mode])
                _LOG.debug('%s', dump)
                dump = dump.replace('\n', '').replace(' ', '')
                self.assertEqual(dump, example['dumps'][mode], msg=(description, mode))

    def test_dump_files_comparison(self):
        """Print the same data as other existing modules."""
        for path in PATHS:
            with open(path, 'r', encoding='utf-8') as py_file:
                code = py_file.read()

            untyped_tree = ast.parse(source=code, filename=path)
            untyped_dump = astunparse.dump(untyped_tree)
            tested_untyped_dump = typed_astunparse.dump(untyped_tree)

            self.assertEqual(untyped_dump.splitlines(), tested_untyped_dump.splitlines())

            typed_tree = typed_ast.ast3.parse(source=code, filename=path)
            bad_typed_dump = astunparse.dump(typed_tree)

            for annotate_fields in [True, False]:
                for include_attributes in [False, True]:
                    if include_attributes and not annotate_fields:
                        continue  # behaviour differs from typed_ast

                    with self.assertRaises(TypeError):
                        _ = typed_ast.ast3.dump(
                            untyped_tree, annotate_fields=annotate_fields,
                            include_attributes=include_attributes)

                    typed_dump = typed_ast.ast3.dump(
                        typed_tree, annotate_fields=annotate_fields,
                        include_attributes=include_attributes)
                    tested_typed_dump = _postprocess_dump(typed_astunparse.dump(
                        typed_tree, annotate_fields=annotate_fields,
                        include_attributes=include_attributes))

                    if include_attributes:
                        # because of https://github.com/python/typed_ast/issues/23
                        self.assertEqual(
                            typed_dump.replace(' ', ''), tested_typed_dump.replace(' ', ''))
                        continue
                    self.assertNotEqual(untyped_dump, bad_typed_dump)
                    self.assertNotEqual(typed_dump, bad_typed_dump)
                    self.assertEqual(typed_dump, tested_typed_dump)

    def test_many_dump_roundtrips(self):
        """Preserve ASTs after unparse(parse(...unparse(parse(dump(tree)))...))."""
        for description, example in EXAMPLES.items():
            for mode in MODES:
                if example['trees'][mode] is None:
                    continue

                dump = typed_astunparse.dump(example['trees'][mode])
                for _ in range(4):
                    tree = typed_ast.ast3.parse(source=dump, mode=mode)
                    dump = typed_astunparse.unparse(tree)
                    _LOG.debug('%s', dump)
                    clean_dump = dump.replace('\n', '').replace(' ', '')
                    self.assertEqual(clean_dump, example['dumps'][mode], msg=(description, mode))
                # TODO: use tree equality comparison below
                # tree = typed_ast.ast3.parse(source=dump, mode=mode)
                # self.assertTrue(typed_astunparse.equal(tree, example['trees'][mode]))



"""Class: Printer."""

import ast
import sys

import astunparse
import typed_ast.ast3


class Printer(astunparse.Printer):
    """Partial rewrite of Printer from astunparse to handle typed_ast.ast3-based trees."""

    def __init__(
            self, file=sys.stdout, indent="  ", annotate_fields: bool = True,
            include_attributes: bool = False):
        """Initialize Printer instance."""
        super().__init__(file=file, indent=indent)
        self._annotate_fields = annotate_fields
        self._include_attributes = include_attributes

    def _prepare_for_print(self, node):
        if isinstance(node, list):
            nodestart = "["
            nodeend = "]"
            children = [("", child) for child in node]
        else:
            nodestart = type(node).__name__ + "("
            nodeend = ")"
            children = [
                (name + "=" if self._annotate_fields else '', value)
                for name, value in typed_ast.ast3.iter_fields(node)]
            if self._include_attributes and node._attributes:
                children += [
                    (attr + '=' if self._annotate_fields else '', getattr(node, attr))
                    for attr in node._attributes]

        return nodestart, children, nodeend

    def generic_visit(self, node):
        """Print the syntax tree without unparsing it.

        Merge of astunparse.Printer.generic_visit() and typed_ast.ast3.dump().
        """
        nodestart, children, nodeend = self._prepare_for_print(node)

        if len(children) > 1:
            self.indentation += 1

        self.write(nodestart)
        for i, pair in enumerate(children):
            attr, child = pair
            if len(children) > 1:
                self.write("\n" + self.indent_with * self.indentation)
            if isinstance(child, (ast.AST, typed_ast.ast3.AST, list)):
                self.write(attr)
                self.visit(child)
            else:
                self.write(attr + repr(child))

            if i != len(children) - 1:
                self.write(",")
        self.write(nodeend)

        if len(children) > 1:
            self.indentation -= 1



"""Class: Unparser."""

import ast

import astunparse
from astunparse.unparser import interleave
import typed_ast.ast3


class Unparser(astunparse.Unparser):
    """Partial rewrite of Unparser from astunparse to handle typed_ast.ast3-based trees.

    The unparser aims at compatibility with native AST, as well as typed AST.

    Quoting grammar file[1] used by CPython, below is copy of the file for Python 3.6:

    "
    -- ASDL's 7 builtin types are:
    -- identifier, int, string, bytes, object, singleton, constant
    --
    -- singleton: None, True or False
    -- constant can be None, whereas None means "no value" for object.

    module Python
    {
        mod = Module(stmt* body)
            | Interactive(stmt* body)
            | Expression(expr body)

            -- not really an actual node but useful in Jython's typesystem.
            | Suite(stmt* body)

        stmt = FunctionDef(identifier name, arguments args,
                           stmt* body, expr* decorator_list, expr? returns)
             | AsyncFunctionDef(identifier name, arguments args,
                                stmt* body, expr* decorator_list, expr? returns)

             | ClassDef(identifier name, expr* bases, keyword* keywords,
                        stmt* body, expr* decorator_list)
             | Return(expr? value)

             | Delete(expr* targets)
             | Assign(expr* targets, expr value)
             | AugAssign(expr target, operator op, expr value)
              -- 'simple' indicates that we annotate simple name without parens
             | AnnAssign(expr target, expr annotation, expr? value, int simple)

              -- use 'orelse' because else is a keyword in target languages
             | For(expr target, expr iter, stmt* body, stmt* orelse)
             | AsyncFor(expr target, expr iter, stmt* body, stmt* orelse)
             | While(expr test, stmt* body, stmt* orelse)
             | If(expr test, stmt* body, stmt* orelse)
             | With(withitem* items, stmt* body)
             | AsyncWith(withitem* items, stmt* body)

             | Raise(expr? exc, expr? cause)
             | Try(stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody)
             | Assert(expr test, expr? msg)

             | Import(alias* names)
             | ImportFrom(identifier? module, alias* names, int? level)

             | Global(identifier* names)
             | Nonlocal(identifier* names)
             | Expr(expr value)
             | Pass | Break | Continue

              -- col_offset is the byte offset in the utf8 string the parser uses
              attributes (int lineno, int col_offset)

              -- BoolOp() can use left & right?
        expr = BoolOp(boolop op, expr* values)
             | BinOp(expr left, operator op, expr right)
             | UnaryOp(unaryop op, expr operand)
             | Lambda(arguments args, expr body)
             | IfExp(expr test, expr body, expr orelse)
             | Dict(expr* keys, expr* values)
             | Set(expr* elts)
             | ListComp(expr elt, comprehension* generators)
             | SetComp(expr elt, comprehension* generators)
             | DictComp(expr key, expr value, comprehension* generators)
             | GeneratorExp(expr elt, comprehension* generators)
             -- the grammar constrains where yield expressions can occur
             | Await(expr value)
             | Yield(expr? value)
             | YieldFrom(expr value)
             -- need sequences for compare to distinguish between
             -- x < 4 < 3 and (x < 4) < 3
             | Compare(expr left, cmpop* ops, expr* comparators)
             | Call(expr func, expr* args, keyword* keywords)
             | Num(object n) -- a number as a PyObject.
             | Str(string s) -- need to specify raw, unicode, etc?
             | FormattedValue(expr value, int? conversion, expr? format_spec)
             | JoinedStr(expr* values)
             | Bytes(bytes s)
             | NameConstant(singleton value)
             | Ellipsis
             | Constant(constant value)

             -- the following expression can appear in assignment context
             | Attribute(expr value, identifier attr, expr_context ctx)
             | Subscript(expr value, slice slice, expr_context ctx)
             | Starred(expr value, expr_context ctx)
             | Name(identifier id, expr_context ctx)
             | List(expr* elts, expr_context ctx)
             | Tuple(expr* elts, expr_context ctx)

              -- col_offset is the byte offset in the utf8 string the parser uses
              attributes (int lineno, int col_offset)

        expr_context = Load | Store | Del | AugLoad | AugStore | Param

        slice = Slice(expr? lower, expr? upper, expr? step)
              | ExtSlice(slice* dims)
              | Index(expr value)

        boolop = And | Or

        operator = Add | Sub | Mult | MatMult | Div | Mod | Pow | LShift
                     | RShift | BitOr | BitXor | BitAnd | FloorDiv

        unaryop = Invert | Not | UAdd | USub

        cmpop = Eq | NotEq | Lt | LtE | Gt | GtE | Is | IsNot | In | NotIn

        comprehension = (expr target, expr iter, expr* ifs, int is_async)

        excepthandler = ExceptHandler(expr? type, identifier? name, stmt* body)
                        attributes (int lineno, int col_offset)

        arguments = (arg* args, arg? vararg, arg* kwonlyargs, expr* kw_defaults,
                     arg? kwarg, expr* defaults)

        arg = (identifier arg, expr? annotation)
               attributes (int lineno, int col_offset)

        -- keyword arguments supplied to call (NULL identifier for **kwargs)
        keyword = (identifier? arg, expr value)

        -- import name with optional 'as' alias.
        alias = (identifier name, identifier? asname)

        withitem = (expr context_expr, expr? optional_vars)
    }
    "

    Quoting the current docstring of "typed_ast/ast3.py"[2] from typed-ast repository:

    "
    The `ast3` module helps Python applications to process trees of the Python
    abstract syntax grammar.  The abstract syntax itself might change with
    each Python release; this module helps to find out programmatically what
    the current grammar looks like and allows modifications of it.  The
    difference between the `ast3` module and the builtin `ast` module is
    that `ast3` is version-independent and provides PEP 484 type comments as
    part of the AST.

    Specifically, these changes are made to the latest Python 3 AST:
    - The `FunctionDef`, `AsyncFunctionDef`, `Assign`, `For`, `AsyncFor`,
      `With`, `AsyncWith`, and `arg` classes all have a `type_comment` field
      which contains a `str` with the text of the associated type comment, if
      any.
    - `parse` has been augmented so it can parse function signature types when
      called with `mode=func_type`.
    - `parse` has an additional argument `feature_version`, which disables
      newer Python syntax features.
    - `Module` has a `type_ignores` field which contains a list of
      lines which have been `# type: ignore`d.
    - `Str` has a `kind` string field which preserves the original string
      prefix, so that `ast3.parse('u"test"').body[0].value.kind == 'u'`.

    An abstract syntax tree can be generated by using the `parse()`
    function from this module.  The result will be a tree of objects whose
    classes all inherit from `ast3.AST`.

    Additionally various helper functions are provided that make working with
    the trees simpler.  The main intention of the helper functions and this
    module in general is to provide an easy to use interface for libraries
    that work tightly with the python syntax (template engines for example).
    "

    [1]: https://hg.python.org/cpython/file/tip/Parser/Python.asdl
    [2]: https://github.com/python/typed_ast/blob/master/typed_ast/ast3.py#L5

    """

    boolops = {typed_ast.ast3.And: 'and', typed_ast.ast3.Or: 'or'}
    """Mapping from boolean operation node to its string representation.

    This overrides of base class dict, because {ast.And: 'and', ast.Or: 'or'} obviously causes
    errors.
    """

    boolops.update(astunparse.Unparser.boolops)

    def _write_string_or_dispatch(self, value):
        """If value is str, write it. Otherwise, dispatch it."""
        if isinstance(value, str):
            self.write(value)
        else:
            self.dispatch(value)

    def _fill_type_comment(self, type_comment):
        """Unparse type comment, adding it on the next line."""
        self.fill('# type: ')
        self._write_string_or_dispatch(type_comment)

    def _write_type_comment(self, type_comment):
        """Unparse type comment, appending it to the end of the current line."""
        self.write('  # type: ')
        self._write_string_or_dispatch(type_comment)

    def _write_raw_literal(self, text: str):
        delimiter = None
        for _ in ("'", '"', "'''", '"""'):
            if _ not in text:
                delimiter = _
                break
        if delimiter is None:
            delimiter = '"""'
        assert delimiter is not None
        if '\n' in text and delimiter in {'"', "'"}:
            delimiter = {'"': '"""', "'": "'''"}[delimiter]
        if delimiter in text:
            escaped_delimiter = ''.join(['\\{}'.format(_) for _ in delimiter])
            text = text.replace(delimiter, escaped_delimiter)
        self.write(delimiter)
        self.write(text)
        self.write(delimiter)

    def _ClassDef(self, t):
        if isinstance(t, ast.ClassDef):
            super()._ClassDef(t)
            return

        self.write("\n")
        for deco in t.decorator_list:
            self.fill("@")
            self.dispatch(deco)
        self.fill("class " + t.name)
        self.write("(")
        comma = False
        for base in t.bases:
            if comma:
                self.write(", ")
            else:
                comma = True
            self.dispatch(base)
        for keyword in t.keywords:
            if comma:
                self.write(", ")
            else:
                comma = True
            self.dispatch(keyword)
        self.write(")")
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _generic_FunctionDef(self, t, async_=False):
        """Unparse FunctionDef or AsyncFunctionDef node.

        Rather than handling:

        FunctionDef/AsyncFunctionDef(
            identifier name, arguments args, stmt* body, expr* decorator_list, expr? returns)

        handle:

        FunctionDef/AsyncFunctionDef(
            identifier name, arguments args, stmt* body, expr* decorator_list, expr? returns,
            string? type_comment)
        """
        if not hasattr(t, 'type_comment') or t.type_comment is None:
            super()._generic_FunctionDef(t, async_)
            return

        self.write("\n")
        for deco in t.decorator_list:
            self.fill("@")
            self.dispatch(deco)
        self.fill(("async " if async_ else "") + "def " + t.name + "(")
        self.dispatch(t.args)
        self.write(")")
        if getattr(t, "returns", False):
            self.write(" -> ")
            self.dispatch(t.returns)
        self.enter()
        self._fill_type_comment(t.type_comment)
        self.dispatch(t.body)
        self.leave()

    def _Assign(self, t):
        """Unparse Assign node.

        Rather than handling just:

        Assign(expr* targets, expr value)

        handle:

        Assign(expr* targets, expr value, string? type_comment)
        """
        super()._Assign(t)
        if hasattr(t, 'type_comment') and t.type_comment is not None:
            self._write_type_comment(t.type_comment)

    def _generic_For(self, t, async_=False):
        """Unparse For or AsyncFor node.

        Rather than handling just:

        For/AsyncFor(expr target, expr iter, stmt* body, stmt* orelse)

        handle:

        For/AsyncFor(expr target, expr iter, stmt* body, stmt* orelse, string? type_comment)
        """
        if not hasattr(t, 'type_comment') or t.type_comment is None:
            super()._generic_For(t, async_)
            return

        self.fill("async for " if async_ else "for ")
        self.dispatch(t.target)
        self.write(" in ")
        self.dispatch(t.iter)
        self.enter()
        self._write_type_comment(t.type_comment)
        self.dispatch(t.body)
        self.leave()
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()

    def _If(self, t):
        self.fill("if ")
        self.dispatch(t.test)
        self.enter()
        self.dispatch(t.body)
        self.leave()
        # collapse nested ifs into equivalent elifs.
        while t.orelse and len(t.orelse) == 1 \
                and isinstance(t.orelse[0], (ast.If, typed_ast.ast3.If)):
            t = t.orelse[0]
            self.fill("elif ")
            self.dispatch(t.test)
            self.enter()
            self.dispatch(t.body)
            self.leave()
        # final else
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()

    def _generic_With(self, t, async_=False):
        """Unparse With or AsyncWith node.

        Rather than handling just:

        With/AsyncWith(withitem* items, stmt* body)

        handle:

        With/AsyncWith(withitem* items, stmt* body, string? type_comment)
        """
        if not hasattr(t, 'type_comment') or t.type_comment is None:
            super()._generic_With(t, async_)
            return

        self.fill("async with " if async_ else "with ")
        interleave(lambda: self.write(", "), self.dispatch, t.items)
        self.enter()
        self._write_type_comment(t.type_comment)
        self.dispatch(t.body)
        self.leave()

    def _Bytes(self, tree):
        if hasattr(tree, 'kind') and tree.kind:
            self.write(tree.kind)
            if 'r' in tree.kind.lower():
                self._write_raw_literal(tree.s.decode())
            else:
                self.write(repr(tree.s)[1:])
            return
        super()._Bytes(tree)

    def _Str(self, tree):
        if hasattr(tree, 'kind') and tree.kind:
            self.write(tree.kind)
            if 'r' in tree.kind.lower():
                self._write_raw_literal(tree.s)
                return
        super()._Str(tree)

    def _Attribute(self, t):
        self.dispatch(t.value)
        # Special case: 3.__abs__() is a syntax error, so if t.value
        # is an integer literal then we need to either parenthesize
        # it or add an extra space to get 3 .__abs__().
        if isinstance(t.value, (ast.Num, typed_ast.ast3.Num)) and isinstance(t.value.n, int):
            self.write(" ")
        self.write(".")
        self.write(t.attr)

    def _Call(self, t):
        if isinstance(t, ast.Call):
            super()._Call(t)
            return

        self.dispatch(t.func)
        self.write("(")
        comma = False
        for arg in t.args:
            if comma:
                self.write(", ")
            else:
                comma = True
            self.dispatch(arg)
        for keyword in t.keywords:
            if comma:
                self.write(", ")
            else:
                comma = True
            self.dispatch(keyword)
        self.write(")")

    def _arguments(self, t):
        first = True
        latest_comment = None
        # normal arguments
        defaults = [None] * (len(t.args) - len(t.defaults)) + t.defaults
        for arg, default in zip(t.args, defaults):
            if first:
                first = False
            else:
                self.write(',')
                if latest_comment is not None:
                    self._write_type_comment(latest_comment)
                    self.fill('        ')
                    latest_comment = None
                else:
                    self.write(' ')
            self.dispatch(arg)
            if default:
                self.write("=")
                self.dispatch(default)
            latest_comment = getattr(arg, 'type_comment', None)

        # varargs, or bare '*' if no varargs but keyword-only arguments present
        if t.vararg or getattr(t, "kwonlyargs", False):
            if first:
                first = False
            else:
                self.write(',')
                if latest_comment is not None:
                    self._write_type_comment(latest_comment)
                    self.fill('        ')
                    latest_comment = None
                else:
                    self.write(' ')
            self.write("*")
            if t.vararg:
                self.write(t.vararg.arg)
                if t.vararg.annotation:
                    self.write(": ")
                    self.dispatch(t.vararg.annotation)
                latest_comment = getattr(t.vararg, 'type_comment', None)

        # keyword-only arguments
        if getattr(t, "kwonlyargs", False):
            for kwarg, default in zip(t.kwonlyargs, t.kw_defaults):
                self.write(',')
                if latest_comment is not None:
                    self._write_type_comment(latest_comment)
                    self.fill('        ')
                    latest_comment = None
                else:
                    self.write(' ')
                self.dispatch(kwarg)
                if default:
                    self.write("=")
                    self.dispatch(default)
                latest_comment = getattr(kwarg, 'type_comment', None)

        # kwargs
        if t.kwarg:
            if first:
                first = False
            else:
                self.write(',')
                if latest_comment is not None:
                    self._write_type_comment(latest_comment)
                    self.fill('        ')
                    latest_comment = None
                else:
                    self.write(' ')
            self.write("**"+t.kwarg.arg)
            if t.kwarg.annotation:
                self.write(": ")
                self.dispatch(t.kwarg.annotation)
            latest_comment = getattr(t.kwarg, 'type_comment', None)

        if latest_comment is not None:
            self._write_type_comment(latest_comment)
            self.fill('        ')



"""Tested function: unparse."""

import ast
import itertools
import logging
import pathlib
import unittest
import sys

import typed_ast.ast3
import typed_astunparse

from .examples import MODES, EXAMPLES, UNVERIFIED_EXAMPLES, INVALID_EXAMPLES, PATHS

_LOG = logging.getLogger(__name__)


class UnparseTests(unittest.TestCase):

    """Unit tests for unparse() function."""

    maxDiff = None

    def test_unparse_examples(self):
        """Unparse ASTs of examples correctly."""
        for description, example in itertools.chain(EXAMPLES.items(), UNVERIFIED_EXAMPLES.items()):
            for mode in MODES:
                if example['trees'][mode] is None:
                    continue
                with self.subTest(description=description):
                    code = typed_astunparse.unparse(example['trees'][mode])
                    _LOG.debug('%s', code)
                    code = code.strip()
                    self.assertEqual(code, example['code'], msg=(description, mode))

    def test_unparse_invalid_examples(self):
        """Raise errors on ASTs of invalid examples as expected."""
        for description, example in INVALID_EXAMPLES.items():
            for mode in MODES:
                if example['trees'][mode] is None:
                    continue
                with self.assertRaises(SyntaxError, msg=(description, mode)):
                    typed_ast.ast3.parse(source=example['code'], mode=mode)

                code = typed_astunparse.unparse(example['trees'][mode])
                tree = None
                try:
                    tree = typed_ast.ast3.parse(source=code, mode=mode)
                except SyntaxError:
                    continue
                code = typed_astunparse.unparse(tree)

    def test_bad_raw_literal(self):
        raw_literal = rb'''\t\t ' """ ''' + rb""" " ''' \n"""
        tree = typed_ast.ast3.Bytes(raw_literal, 'rb')
        # with self.assertRaises(SyntaxError):
        code = typed_astunparse.unparse(tree)
        print(code)
        for mode in MODES:
            tree = typed_ast.ast3.parse(source=code, mode=mode)

    def test_many_roundtrips(self):
        """Prserve ASTs when doing parse(unparse(parse(...unparse(parse(code))...)))."""
        for description, example in EXAMPLES.items():
            for mode in MODES:
                if example['trees'][mode] is None:
                    continue

                tree = example['trees'][mode]
                with self.subTest(description=description):
                    for _ in range(4):
                        code = typed_astunparse.unparse(tree)
                        _LOG.debug('%s', code)
                        clean_code = code.strip()
                        self.assertEqual(clean_code, example['code'], msg=(description, mode))
                        tree = typed_ast.ast3.parse(source=code, mode=mode)

    def test_files(self):
        """Keep Python stdlib tree the same after roundtrip parse-unparse."""
        for path in PATHS:
            if sys.version_info[:2] == (3, 7) and pathlib.Path(path).name == 'dataclasses.py':
                continue
            with open(path, 'r', encoding='utf-8') as py_file:
                original_code = py_file.read()
            tree = typed_ast.ast3.parse(source=original_code, filename=path)
            code = typed_astunparse.unparse(tree)
            with self.subTest(path=path):
                roundtrip_tree = typed_ast.ast3.parse(source=code)
                tree_dump = typed_ast.ast3.dump(tree, include_attributes=False)
                roundtrip_tree_dump = typed_ast.ast3.dump(roundtrip_tree, include_attributes=False)
                self.assertEqual(tree_dump, roundtrip_tree_dump, msg=path)

    def test_untyped_files(self):
        """Unparse Python stdlib correctly even if parsed using built-in ast package."""
        for path in PATHS:
            with open(path, 'r', encoding='utf-8') as py_file:
                original_code = py_file.read()
            tree = ast.parse(source=original_code, filename=path)
            code = typed_astunparse.unparse(tree)
            with self.subTest(path=path):
                roundtrip_tree = ast.parse(source=code)
                tree_dump = ast.dump(tree, include_attributes=False)
                roundtrip_tree_dump = ast.dump(roundtrip_tree, include_attributes=False)
                self.assertEqual(tree_dump, roundtrip_tree_dump, msg=path)



"""This is setup.py file for typed-astunparse."""

import setup_boilerplate


class Package(setup_boilerplate.Package):

    """Package metadata."""

    name = 'typed-astunparse'
    description = 'typed-astunparse is to typed-ast as astunparse is to ast'
    download_url = 'https://github.com/mbdevpl/typed-astunparse'
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Pre-processors',
        'Topic :: Utilities']
    keywords = ['ast', 'unparsing', 'pretty printing']


if __name__ == '__main__':
    Package.setup()



"""Below code is generic boilerplate and normally should not be changed.

To avoid setup script boilerplate, create "setup.py" file with the following minimal contents
and modify them according to the specifics of your package.

See the implementation of setup_boilerplate.Package for default metadata values and available
options.
"""

import importlib
import pathlib
import sys
import typing as t

import docutils.nodes
import docutils.parsers.rst
import docutils.utils
import setuptools

__updated__ = '2018-02-14'

SETUP_TEMPLATE = '''"""Setup script."""

import setup_boilerplate


class Package(setup_boilerplate.Package):

    """Package metadata."""

    name = ''
    description = ''
    download_url = 'https://github.com/mbdevpl/...'
    classifiers = [
        'Development Status :: 1 - Planning',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only']
    keywords = []


if __name__ == '__main__':
    Package.setup()
'''

HERE = pathlib.Path(__file__).resolve().parent


def find_version(
        package_name: str, version_module_name: str = '_version',
        version_variable_name: str = 'VERSION') -> str:
    """Simulate behaviour of "from package_name._version import VERSION", and return VERSION."""
    version_module = importlib.import_module(
        '{}.{}'.format(package_name.replace('-', '_'), version_module_name))
    return getattr(version_module, version_variable_name)


def find_packages(root_directory: str = '.') -> t.List[str]:
    """Find packages to pack."""
    exclude = ['test', 'test.*'] if ('bdist_wheel' in sys.argv or 'bdist' in sys.argv) else []
    packages_list = setuptools.find_packages(root_directory, exclude=exclude)
    return packages_list


def parse_requirements(
        requirements_path: str = 'requirements.txt') -> t.List[str]:
    """Read contents of requirements.txt file and return data from its relevant lines.

    Only non-empty and non-comment lines are relevant.
    """
    requirements = []
    with open(str(HERE.joinpath(requirements_path))) as reqs_file:
        for requirement in [line.strip() for line in reqs_file.read().splitlines()]:
            if not requirement or requirement.startswith('#'):
                continue
            requirements.append(requirement)
    return requirements


def partition_version_classifiers(
        classifiers: t.Sequence[str], version_prefix: str = 'Programming Language :: Python :: ',
        only_suffix: str = ' :: Only') -> t.Tuple[t.List[str], t.List[str]]:
    """Find version number classifiers in given list and partition them into 2 groups."""
    versions_min, versions_only = [], []
    for classifier in classifiers:
        version = classifier.replace(version_prefix, '')
        versions = versions_min
        if version.endswith(only_suffix):
            version = version.replace(only_suffix, '')
            versions = versions_only
        try:
            versions.append(tuple([int(_) for _ in version.split('.')]))
        except ValueError:
            pass
    return versions_min, versions_only


def find_required_python_version(
        classifiers: t.Sequence[str], version_prefix: str = 'Programming Language :: Python :: ',
        only_suffix: str = ' :: Only') -> t.Optional[str]:
    """Determine the minimum required Python version."""
    versions_min, versions_only = partition_version_classifiers(
        classifiers, version_prefix, only_suffix)
    if len(versions_only) > 1:
        raise ValueError(
            'more than one "{}" version encountered in {}'.format(only_suffix, versions_only))
    only_version = None
    if len(versions_only) == 1:
        only_version = versions_only[0]
        for version in versions_min:
            if version[:len(only_version)] != only_version:
                raise ValueError(
                    'the "{}" version {} is inconsistent with version {}'
                    .format(only_suffix, only_version, version))
    min_supported_version = None
    for version in versions_min:
        if min_supported_version is None or \
                (len(version) >= len(min_supported_version) and version < min_supported_version):
            min_supported_version = version
    if min_supported_version is None:
        if only_version is not None:
            return '.'.join([str(_) for _ in only_version])
    else:
        return '>=' + '.'.join([str(_) for _ in min_supported_version])
    return None


def parse_rst(text: str) -> docutils.nodes.document:
    """Parse text assuming it's an RST markup."""
    parser = docutils.parsers.rst.Parser()
    components = (docutils.parsers.rst.Parser,)
    settings = docutils.frontend.OptionParser(components=components).get_default_values()
    document = docutils.utils.new_document('<rst-doc>', settings=settings)
    parser.parse(text, document)
    return document


class SimpleRefCounter(docutils.nodes.NodeVisitor):

    """Find all simple references in a given docutils document."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.references = []

    def visit_reference(self, node: docutils.nodes.reference) -> None:
        """Called for "reference" nodes."""
        if len(node.children) != 1 or not isinstance(node.children[0], docutils.nodes.Text) \
                or not all(_ in node.attributes for _ in ('name', 'refuri')):
            return
        path = pathlib.Path(node.attributes['refuri'])
        try:
            if path.is_absolute():
                return
            resolved_path = path.resolve()
        except FileNotFoundError:  # in resolve(), prior to Python 3.6
            return
        except OSError:  # in is_absolute() and resolve(), on URLs in Windows
            return
        try:
            resolved_path.relative_to(HERE)
        except ValueError:
            return
        if not path.is_file():
            return
        assert node.attributes['name'] == node.children[0].astext()
        self.references.append(node)

    def unknown_visit(self, node: docutils.nodes.Node) -> None:
        """Called for unknown node types."""
        pass


def resolve_relative_rst_links(text: str, base_link: str):
    """Resolve all relative links in a given RST document.

    All links of form `link`_ become `link <base_link/link>`_.
    """
    document = parse_rst(text)
    visitor = SimpleRefCounter(document)
    document.walk(visitor)
    for target in visitor.references:
        name = target.attributes['name']
        uri = target.attributes['refuri']
        new_link = '`{} <{}{}>`_'.format(name, base_link, uri)
        if name == uri:
            text = text.replace('`<{}>`_'.format(uri), new_link)
        else:
            text = text.replace('`{} <{}>`_'.format(name, uri), new_link)
    return text


class Package:

    """Default metadata and behaviour for a Python package setup script."""

    root_directory = '.'  # type: str
    """Root directory of the source code of the package, relative to the setup.py file location."""

    name = None  # type: str

    version = None  # type: str
    """"If None, it will be obtained from "package_name._version.VERSION" variable."""

    description = None  # type: str

    long_description = None  # type: str
    """If None, it will be generated from readme."""

    url = 'https://mbdevpl.github.io/'  # type: str
    download_url = 'https://github.com/mbdevpl'  # type: str
    author = 'Mateusz Bysiek'  # type: str
    author_email = 'mb@mbdev.pl'  # type: str
    # maintainer = None  # type: str
    # maintainer_email = None  # type: str
    license_str = 'Apache License 2.0'  # type: str

    classifiers = []  # type: t.List[str]
    """List of valid project classifiers: https://pypi.python.org/pypi?:action=list_classifiers"""

    keywords = []  # type: t.List[str]

    packages = None  # type: t.List[str]
    """If None, determined with help of setuptools."""

    package_data = {}
    exclude_package_data = {}

    install_requires = None  # type: t.List[str]
    """If None, determined using requirements.txt."""

    extras_require = {}  # type: t.Mapping[str, t.List[str]]
    """A dictionary containing entries of type 'some_feature': ['requirement1', 'requirement2']."""

    python_requires = None  # type: str
    """If None, determined from provided classifiers."""

    entry_points = {}  # type: t.Mapping[str, t.List[str]]
    """A dictionary used to enable automatic creation of console scripts, gui scripts and plugins.

    Example entry:
    'console_scripts': ['script_name = package.subpackage:function']
    """

    test_suite = 'test'  # type: str

    @classmethod
    def try_fields(cls, *names) -> t.Optional[t.Any]:
        """Return first existing of given class field names."""
        for name in names:
            if hasattr(cls, name):
                return getattr(cls, name)
        raise AttributeError((cls, names))

    @classmethod
    def parse_readme(cls, readme_path: str = 'README.rst', encoding: str = 'utf-8') -> str:
        """Parse readme and resolve relative links in it if it is feasible.

        Links are resolved if readme is in rst format and the package is hosted on GitHub.
        """
        with open(str(HERE.joinpath(readme_path)), encoding=encoding) as readme_file:
            long_description = readme_file.read()  # type: str

        if readme_path.endswith('.rst') and cls.download_url.startswith('https://github.com/'):
            base_url = '{}/blob/v{}/'.format(cls.download_url, cls.version)
            long_description = resolve_relative_rst_links(long_description, base_url)

        return long_description

    @classmethod
    def prepare(cls) -> None:
        """Fill in possibly missing package metadata."""
        if cls.version is None:
            cls.version = find_version(cls.name)
        if cls.long_description is None:
            cls.long_description = cls.parse_readme()
        if cls.packages is None:
            cls.packages = find_packages(cls.root_directory)
        if cls.install_requires is None:
            cls.install_requires = parse_requirements()
        if cls.python_requires is None:
            cls.python_requires = find_required_python_version(cls.classifiers)

    @classmethod
    def setup(cls) -> None:
        """Run setuptools.setup() with correct arguments."""
        cls.prepare()
        setuptools.setup(
            name=cls.name, version=cls.version, description=cls.description,
            long_description=cls.long_description, url=cls.url, download_url=cls.download_url,
            author=cls.author, author_email=cls.author_email,
            maintainer=cls.try_fields('maintainer', 'author'),
            maintainer_email=cls.try_fields('maintainer_email', 'author_email'),
            license=cls.license_str, classifiers=cls.classifiers, keywords=cls.keywords,
            packages=cls.packages, package_dir={'': cls.root_directory},
            include_package_data=True,
            package_data=cls.package_data, exclude_package_data=cls.exclude_package_data,
            install_requires=cls.install_requires, extras_require=cls.extras_require,
            python_requires=cls.python_requires,
            entry_points=cls.entry_points, test_suite=cls.test_suite)



"""This is "__init__.py" file for "typed_astunparse" package.

functions: unparse, dump
"""

import ast
import typing as t

import typed_ast.ast3
from six.moves import cStringIO

from .unparser import Unparser
from .printer import Printer
from ._version import VERSION

__version__ = VERSION


def unparse(tree: t.Union[ast.AST, typed_ast.ast3.AST]) -> str:
    """Unparse the abstract syntax tree into a str.

    Behave just like astunparse.unparse(tree), but handle trees which are typed, untyped, or mixed.
    In other words, a mixture of ast.AST-based and typed_ast.ast3-based nodes will be unparsed.
    """
    stream = cStringIO()
    Unparser(tree, file=stream)
    return stream.getvalue()


def dump(
        tree: t.Union[ast.AST, typed_ast.ast3.AST], annotate_fields: bool = True,
        include_attributes: bool = False) -> str:
    """Behave just like astunparse.dump(tree), but handle typed_ast.ast3-based trees."""
    stream = cStringIO()
    Printer(
        file=stream, annotate_fields=annotate_fields,
        include_attributes=include_attributes).visit(tree)
    return stream.getvalue()


__all__ = ['unparse', 'dump']
